@@ -7,6 +7,7 @@ import pickle as pkl
 import logging
 import time
 import numpy as np
+from pathlib import Path
 
 from cognifly.utils.udp_interface import UDPInterface
 from cognifly.utils.tcp_video_interface import TCPVideoInterface
@@ -28,9 +29,6 @@ EASY_API_MAX_ALTITUDE = 0.9
 EASY_API_MIN_ALTITUDE = 0.35
 
 
-# Global GUI thread (only one GUI is available for all drones)
-
-
 def _gui_process(name, recv_q, send_q):
     """
     Process in charge of managing the Graphical User Interface
@@ -45,14 +43,21 @@ def _gui_process(name, recv_q, send_q):
     import PySimpleGUI as sg
     import cv2
 
+    sprites_folder = Path(__file__).parent.absolute() / 'sprites'
+    sprite_gamepad_on = str(sprites_folder / 'gamepad_on.png')
+    sprite_gamepad_off = str(sprites_folder / 'gamepad_off.png')
+
     sg.theme('BluePurple')
-    layout = [[sg.Text('Battery:'), sg.Text('OK', size=(15, 1), key='-batt-')],
+    layout = [[sg.Text('Battery:'),
+               sg.Text('OK', size=(15, 1), key='-batt-'),
+               sg.Image(filename=sprite_gamepad_off, key='-gamepad-')],
               [sg.Button('DISARM')],
               [sg.Image(filename='', key='-image-')]]
     window = sg.Window(name,
                        layout,
                        keep_on_top=True,
                        enable_close_attempted_event=True)
+    gamepad_elem = window['-gamepad-']
     image_elem = window['-image-']
     batt_elem = window['-batt-']
 
@@ -92,6 +97,10 @@ def _gui_process(name, recv_q, send_q):
                 elif cmd == 'BAT':  # batt str value
                     batt_str = val
                     batt_elem.update(batt_str)
+                elif cmd == 'GMP':  # connected gamepad value
+                    gamepad = val
+                    im_gmp = sprite_gamepad_on if gamepad else sprite_gamepad_off
+                    gamepad_elem.update(im_gmp)
 
             if no_window_event and no_read_event:
                 time.sleep(0.05)
@@ -186,6 +195,7 @@ class Cognifly:
         self._gui_lock = Lock()
         self._gui_display = False
         self._gui_batt_str = "OK"
+        self._gui_gamepad = False
         self._gui_process = None
 
         self._listener_thread = Thread(target=self.__listener_thread)
@@ -225,6 +235,7 @@ class Cognifly:
             self._gui_process.start()
         is_displaying = False
         previous_batt_str = "OK"
+        previous_gamepad = False
         while True:
             send_event = False
             recv_event = False
@@ -262,13 +273,18 @@ class Cognifly:
                 send_q.put(('IMG', None))
                 send_event = True
                 is_displaying = False
-            # battery:
+            # battery and gamepad display:
             with self._gui_lock:
                 batt_str = self._gui_batt_str
+                gamepad = self._gui_gamepad
             if batt_str != previous_batt_str:
                 send_q.put(('BAT', batt_str))
                 send_event = True
                 previous_batt_str = batt_str
+            if gamepad != previous_gamepad:
+                send_q.put(('GMP', gamepad))
+                send_event = True
+                previous_gamepad = gamepad
             # sleep if nothing has occurred:
             if not send_event and not recv_event:
                 time.sleep(0.1)
@@ -292,10 +308,18 @@ class Cognifly:
                 elif m[0] == "OBS":  # observation type
                     self._lock.acquire()
                     i_obs = m[1]
+                    telemetry = m[2]
+                    gamepad = m[3]
+                    voltage = m[4]
+                    debug_flags = m[5]
                     if i_obs > self.__last_i_obs:  # drop outdated observations
-                        self.__obs = m[2]  # observation
+                        self.__obs = (telemetry, gamepad, voltage, debug_flags)  # observation
                         self.__last_i_obs = i_obs
                     self._lock.release()
+                    if gamepad is not None:
+                        with self._gui_lock:
+                            self._gui_batt_str = f"{voltage}V"
+                            self._gui_gamepad = gamepad
                 elif m[0] == "RES":  # acknowledgement for the reset command
                     self._lock.acquire()
                     self.__wait_ack_reset = False
@@ -306,8 +330,9 @@ class Cognifly:
                         self.__wait_done = False
                     self._lock.release()
                 elif m[0] == "BBT":  # bad battery state
-                    with self._gui_lock:
-                        self._gui_batt_str = f"{m[1][0]}V"
+                    pass
+                    # with self._gui_lock:
+                    #     self._gui_batt_str = f"{m[1][0]}V"
 
     def __resender_thread(self):
         """
@@ -326,6 +351,7 @@ class Cognifly:
     def send(self, msg_type, msg=None):
         """
         Send an UDP message to the drone.
+
         Args:
             msg_type: string in ["RES", ACT"].
             msg: tuple of the form (str:command, ...:args).
@@ -379,6 +405,7 @@ class Cognifly:
     def reset(self, sleep_duration=0.1):
         """
         Resets the drone and waits for reset acknowlegement.
+
         Args:
             sleep_duration: float: the reset command waits for at least this amount of time
         """
@@ -412,6 +439,35 @@ class Cognifly:
                     break
             logger.info("observation initialized")
 
+    def set_coordinates(self, x=None, y=None, yaw=None):
+        """
+        Sets the drone current coordinates.
+
+        This computes a new flight origin.
+
+        Args:
+            x: float (optional): new coordinate on the X axis (m)
+            y: float (optional): new coordinate on the Y axis (m)
+            yaw: float (optional): new coordinate on the yaw axis (rad)
+        """
+        self.send(msg_type="SCO", msg=(x, y, yaw))
+
+    def set_flight_origin(self, x=None, y=None, yaw=None):
+        """
+        Sets the drone flight origin directly.
+
+        Use with x=0, y=0 and yaw=0 if you wish the origin to be the same as the pose estimator.
+        This is useful when you are using a custom pose estimator.
+        Note that calling reset() resets the flight origin to the current position of the drone.
+        This also happens on instantiation of the Cognifly object.
+
+        Args:
+            x: float (optional): origin on the X axis (m)
+            y: float (optional): origin on the Y axis (m)
+            yaw: float (optional): origin on the yaw axis (rad)
+        """
+        self.send(msg_type="SFO", msg=(x, y, yaw))
+
     def arm(self):
         """
         Arms the drone.
@@ -424,13 +480,19 @@ class Cognifly:
         """
         self.send(msg_type="ACT", msg=("DISARM",))
 
-    def takeoff_nonblocking(self, altitude=None):
+    def takeoff_nonblocking(self, altitude=None, track_xy=False, max_duration=10.0, max_velocity=0.5):
         """
         The drone takes off.
 
         Do not use altitude (this currently refers to a FC throttle value and is likely to change in the future).
+
+        Args:
+            altitude: do not use
+            track_xy: bool (optional): if True, the drone will track X and Y while taking off
+            max_duration: float (optional): duration for which X and Y can be tracked if track_xy is True (s)
+            max_velocity: float (optional): maximum velocity used to track X and Y if track_xy is True (m/s)
         """
-        self.send(msg_type="ACT", msg=("TAKEOFF", altitude))
+        self.send(msg_type="ACT", msg=("TAKEOFF", altitude, track_xy, max_duration, max_velocity))
         self.time_takeoff = time.time()
 
     def land_nonblocking(self):
@@ -442,7 +504,9 @@ class Cognifly:
     def set_velocity_nonblocking(self, v_x=0.0, v_y=0.0, v_z=0.0, w=0.0, duration=1.0, drone_frame=True):
         """
         Sets the drone velocity.
+
         The drone needs to be armed.
+
         Args:
             v_x: float (optional): velocity target on the frontward axis (m/s); tested range: 0.1 - 0.5
             v_y: float (optional): velocity target on the rightward axis (m/s); tested range: 0.1 - 0.5
@@ -454,25 +518,30 @@ class Cognifly:
         frame_str = "VDF" if drone_frame else "VWF"
         self.send(msg_type="ACT", msg=(frame_str, v_x, v_y, v_z, w, duration))
 
-    def set_position_nonblocking(self, x, y, z, yaw=None, max_velocity=0.5, max_yaw_rate=0.5, max_duration=10.0, relative=False):
+    def set_position_nonblocking(self, x, y, z=None, yaw=None, max_velocity=0.5, max_yaw_rate=0.5, max_duration=10.0, relative=False, relative_z=False):
         """
         Sets a position target for the drone.
         If relative is False (default), the target is relative to the world axis.
         If relative is True, the target is relative to the drone frame.
-        Note: z is always in the world frame.
+        Note: z is always in the world frame, except if both relative and relative_z are True.
         The drone needs to be armed.
+
         Args:
             x: float: x target (m)
             y: float: y target (m)
-            z: float: z target (m); tested range: 0.3 - 0.9
+            z: float (optional): z target (m); tested range: 0.3 - 0.9; when None, z is not tracked
             yaw: float (optional): yaw target (rad); if None, the yaw is not changed
             max_velocity: float (optional): maximum velocity used to go to position (m/s)
             max_yaw_rate: float (optional): maximum yaw rate used to go to yaw (rad/s)
             max_duration: float (optional): maximum duration of the command (s)
             relative: bool (optional): whether to use the drone frame (True) or the world frame (False, default)
+            relative_z: bool (optional): when both relative and relative_z are True, z is in the drone frame
         """
         if relative:
-            frame_str = "PDF"
+            if relative_z:
+                frame_str = "PDZ"
+            else:
+                frame_str = "PDF"
         else:
             frame_str = "PWF"
         self._lock.acquire()
@@ -483,11 +552,13 @@ class Cognifly:
     def set_pid_vel_x(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
         """
         Sets the gains of the PID for the x velocity tracker.
+
         Gains left as None are not affected.
         Example of reasonable values:
             k_p = 750.0
             k_i = 400.0
             k_d = 50.0
+
         Args:
             k_p: float (optional): proportional gain
             k_i: float (optional): integral gain
@@ -500,11 +571,13 @@ class Cognifly:
     def set_pid_vel_y(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
         """
         Sets the gains of the PID for the y velocity tracker.
+
         Gains left as None are not affected.
         Example of reasonable values:
             k_p = 750.0
             k_i = 400.0
             k_d = 50.0
+
         Args:
             k_p: float (optional): proportional gain
             k_i: float (optional): integral gain
@@ -517,11 +590,13 @@ class Cognifly:
     def set_pid_vel_z(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
         """
         Sets the gains of the PID for the z velocity tracker.
+
         Gains left as None are not affected.
         Example of reasonable values:
             k_p = 50.0
             k_i = 20.0
             k_d = 5.0
+
         Args:
             k_p: float (optional): proportional gain
             k_i: float (optional): integral gain
@@ -534,11 +609,13 @@ class Cognifly:
     def set_pid_vel_w(self, k_p=None, k_i=None, k_d=None, timeout=1.0):
         """
         Sets the gains of the PID for the yaw rate tracker.
+
         Gains left as None are not affected.
         Example of reasonable values:
             k_p = 400.0
             k_i = 200.0
             k_d = 0.0
+
         Args:
             k_p: float (optional): proportional gain
             k_i: float (optional): integral gain
@@ -554,20 +631,23 @@ class Cognifly:
     # (don't use along the pro API, otherwise weird things will happen with altitude)
     # (sleeps after calls, uses cm instead of m, and uses degrees instead of rad):
 
-    def takeoff(self, altitude=None, alt_duration=10.0):
+    def takeoff(self, altitude=None, alt_duration=10.0, track_xy=False, wait_before_control=10.0):
         """
         Arms the drone and takes off
+
         Args:
             altitude: float (optional): target altitude (cm)
             alt_duration: float (optional): additional max duration to reach the target altitude
+            track_xy: bool (optional): if True, the drone will track X and Y while taking off
+            wait_before_control: float (optional): time after takeoff before control starts
         """
         altitude = altitude / 100.0 if altitude is not None else EASY_API_TAKEOFF_ALTITUDE
         self.reset()
         self.arm()
         time.sleep(1.0)
         # the takeoff altitude might depend on the battery with this and Z will only be defined properly later:
-        self.takeoff_nonblocking()
-        time.sleep(10.0)
+        self.takeoff_nonblocking(track_xy=track_xy)
+        time.sleep(wait_before_control)
         self.easy_api_cur_z = altitude
         # this works (Z properly defined) but when done from the ground it is a bit violent and not very stable:
         self.set_position_nonblocking(x=0.0, y=0.0, z=self.easy_api_cur_z, yaw=0.0,
@@ -733,6 +813,7 @@ class Cognifly:
     def position_sequence(self, sequence, max_duration=60.0, relative=False, speed=None, yaw_rate=None):
         """
         The drone follows a roadmap defined by a sequence of targets.
+
         Args:
             sequence: a sequence of sequence-like elements, each of length 3 or 4 (mixing 3 and 4 is possible).
                 If the length of an element is 3, it is interpreted as (x, y, z) (cm).
@@ -772,6 +853,7 @@ class Cognifly:
     def curve(self, x1, y1, z1, x2, y2, z2, speed):
         """
         2-positions roadmap.
+
         This is an alias for position_sequence(sequence=[[x1, y1, z1],[x2, y2, z2]], speed=speed).
         """
         self.position_sequence(sequence=[[x1, y1, z1],[x2, y2, z2]], speed=speed)
@@ -779,6 +861,7 @@ class Cognifly:
     def set_speed(self, speed):
         """
         Sets the maximum speed of the drone in the school API
+
         Args:
             speed: float: speed in cm/s; a reasonable value is 20.
         """
@@ -787,6 +870,7 @@ class Cognifly:
     def set_yaw_rate(self, yaw_rate):
         """
         Sets the maximum yaw rate of the drone in the school API
+
         Args:
             yaw_rate: float: yaw rate in deg/s; a reasonable value is 30.
         """
@@ -796,18 +880,22 @@ class Cognifly:
 
     # Methods common to both APIs:
 
-    def streamon(self, resolution="VGA", fps=10, display=False, wait_first_frame=True):
+    def streamon(self, resolution="VGA", fps=10, display=False, wait_first_frame=True, format='jpg', quality=95):
         """
         Starts camera streaming.
+
         CAUTION: This will slow the frequency of the onboard controller down and may make the drone unstable!
         The highest resolution and fps are, the worst the influence on the onboard controller.
         The image transfers happens over TCP, which comes with a noticeable delay.
         CAUTION: High fps may saturate the network in the presence of several drones.
+
         Args:
-            resolution: Tuple(float: height, float: width): resolution of the captured images.
+            resolution: Tuple(float: width, float: height): resolution of the captured images.
             fps: integer: maximum framerate (may not be attained),
             display: boolean: whether to display the stream in the GUI (requires GUI to be enabled),
             wait_first_frame: boolean: whether to sleep until the first frame is available.
+            format: string: one of 'jpg', 'webp' or 'png'
+            quality: int: image quality between 0 and 100
         """
         if not self.tcp_video_int.is_receiver_running():
             free_port = get_free_port(self.recv_port_video)
@@ -817,22 +905,23 @@ class Cognifly:
                 self.recv_port_video = free_port
             self.tcp_video_int.start_receiver(self.recv_port_video)
             time.sleep(1.0)  # sleep a bit so the server starts before the drone tries to connect
-            self.send(msg_type="ST1", msg=(self.local_ip, self.recv_port_video, resolution, fps))
+            self.send(msg_type="ST1", msg=(self.local_ip, self.recv_port_video, resolution, fps, format, quality))
             if wait_first_frame:
                 _ = self.get_frame(wait_new_frame=True)
             self._gui_lock.acquire()
             self._gui_display = display
             self._gui_lock.release()
 
-    def stream(self, resolution="VGA", fps=10):
+    def stream(self, resolution="VGA", fps=10, format='jpg', quality=95):
         """
-        Alias for streamon(resolution, fps=fps, display=True)
+        Alias for streamon(resolution, fps=fps, display=True, format=format, quality=quality)
         """
-        self.streamon(resolution, fps, display=True)
+        self.streamon(resolution, fps, display=True, format=format, quality=quality)
 
     def get_frame(self, wait_new_frame=True, timeout=10.0, sleep_between_trials=0.05):
         """
         Retrieves a single frame from the stream.
+
         Args:
             wait_new_frame: boolean: if True, two consecutive calls to the fonction will not return identical frames.
             timeout: after this duration without retrieving a frame, a TimeoutException will be raised.
@@ -874,18 +963,21 @@ class Cognifly:
     def get_telemetry(self):
         """
         Gets a tuple that describe the state of the drone in the world frame.
+
         Angles are in radians.
+
         Returns:
-            telemetry: Tuple: (battery_voltage: float,
-                               x: float,
+            telemetry: Tuple: (x: float,
                                y: float,
                                z: float,
                                yaw: float,
                                vel_x: float,
                                vel_y: float,
                                vel_z: float,
-                               yaw_rate: float,
-                               health_flags: list of strings)
+                               yaw_rate: float)
+            gamepad_override: Boolean: if True, the telemtry is invalid!
+            voltage: float
+            debug_flags: list of strings
         """
         self._lock.acquire()
         obs = deepcopy(self.__obs)
@@ -894,58 +986,68 @@ class Cognifly:
 
     def get_battery(self):
         """
+        Battery voltage.
+
         Returns:
             battery_voltage: float: battery voltage, in V
         """
-        telemetry = self.get_telemetry()
-        return telemetry[0]
+        return self.get_telemetry()[2]
 
     def get_position(self):
         """
+        *Caution:* invalid when gamepad is in override mode.
+
         Returns:
             x: float: in cm
             y: float: in cm
             z: float: in cm
         """
-        telemetry = self.get_telemetry()
-        return telemetry[1] * 100, telemetry[2] * 100, telemetry[3] * 100
+        telemetry = self.get_telemetry()[0]
+        return telemetry[0] * 100, telemetry[1] * 100, telemetry[2] * 100
 
     def get_yaw(self):
         """
+        *Caution:* invalid when gamepad is in override mode.
+
         Returns:
             x: float: yaw, in deg
         """
-        telemetry = self.get_telemetry()
-        return telemetry[4] * 180.0 / np.pi
+        telemetry = self.get_telemetry()[0]
+        return telemetry[3] * 180.0 / np.pi
 
     def get_velocity(self):
         """
+        *Caution:* invalid when gamepad is in override mode.
+
         Returns:
             v_x: float: in cm/s
             v_y: float: in cm/s
             v_z: float: in cm/s
         """
-        telemetry = self.get_telemetry()
-        return telemetry[5] * 100, telemetry[6] * 100, telemetry[7] * 100
+        telemetry = self.get_telemetry()[0]
+        return telemetry[4] * 100, telemetry[5] * 100, telemetry[6] * 100
 
     def get_health(self):
         """
         Returns:
             health_flags: list os strings: useful flags for debugging
         """
-        telemetry = self.get_telemetry()
-        return telemetry[9]
+        return self.get_telemetry()[3]
 
     def get_yaw_rate(self):
         """
+        *Caution:* invalid when gamepad is in override mode.
+
         Returns:
             w: float: yaw rate in deg/s
         """
-        telemetry = self.get_telemetry()
-        return telemetry[8] * 180.0 / np.pi
+        telemetry = self.get_telemetry()[0]
+        return telemetry[7] * 180.0 / np.pi
 
     def get_tof(self):
         """
+        *Caution:* invalid when gamepad is in override mode.
+
         Returns:
             h: float: altitude in cm
         """
@@ -955,11 +1057,15 @@ class Cognifly:
     def get_height(self):
         """
         Alias for get_tof() as cognifly doesn't know its height from the starting point.
+
+        *Caution:* invalid when gamepad is in override mode.
         """
         return self.get_tof()
 
     def get_speed(self):
         """
+        *Caution:* invalid when gamepad is in override mode.
+
         Returns:
             speed: float: norm of the velocity, in cm/s
         """
